// From the software distribution accompanying the textbook
// "A Practical Introduction to Data Structures and Algorithm Analysis,
// Third Edition (C++)" by Clifford A. Shaffer.
// Source code Copyright (C) 2007-2011 by Clifford A. Shaffer.

// Test for variable nodes implemented with a composite design.
// The test vehicle is an expression tree.
// The program builds and traverses a simple tree.

#include "book.h"

typedef char Operator;
typedef string Operand;

class VarBinNode
{
    public:
        virtual ~VarBinNode() {} // Generic destructor
        virtual bool isLeaf() = 0;
        virtual void traverse(string indent) = 0;
};


// Leaf node
class LeafNode: public VarBinNode
{
    private:
        Operand var; // Operand value

    public:
        LeafNode(const Operand &val)
        {
            var = val;
        } // Constructor

        bool isLeaf()
        {
            return true;
        } // isLeaf for Leafnode

        Operand value()
        {
            return var;
        } // Return node value

        void traverse(string indent)
        {
            cout << indent << "  Leaf: " << value() << endl;
        }
};


// Internal node
class IntlNode: public VarBinNode
{
    private:
        VarBinNode *lc; // Left child
        VarBinNode *rc; // Right child
        Operator opx; // Operator value

    public:
        IntlNode(const Operator &op, VarBinNode *l, VarBinNode *r)
        {
            opx = op;
            lc = l;
            rc = r;
        }

        bool isLeaf()
        {
            return false;
        } //isLeaf for IntlNode

        VarBinNode *left()
        {
            return lc;
        } //Left child

        VarBinNode *right()
        {
            return rc;
        } //Right child

        Operator value()
        {
            return opx;
        } //Value


        // Traversal behavior for internal nodes:
        // backwards in order with indent for depth
        void traverse(string indent)
        {
            if(right() != NULL)
                right()->traverse(indent += "  ");

            cout << indent << "Internal: " << value() << endl;

            if(left() != NULL)
                left()->traverse(indent += "");
        }
};


int main()
{
    // ((5+x)*3)
    VarBinNode *leaf1 = new LeafNode("5");
    VarBinNode *leaf2 = new LeafNode("x");
    VarBinNode *parent1 = new IntlNode('+', leaf1, leaf2);

    VarBinNode *leaf3 = new LeafNode("3");

    VarBinNode *root = new IntlNode('*', parent1, leaf3);

    root->traverse("");

    return 0;
}
